/* 
 * Copyright (c)2007, Dustin Spicuzza
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * The name of the Dustin Spicuzza may not be used to endorse or promote 
 *       products derived from this software without specific prior written 
 *       permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Dustin Spicuzza ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Dustin Spicuzza BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This library extends and supplements the Cypress SuiteUSB .NET C# library
 * 
 * This is not distributed by or affiliated with Cypress or SuiteUSB .NET, and is not
 * supported by them. You need their library to make this work, of course.
 * 
 * Heavily influenced by http://msdn.microsoft.com/msdnmag/issues/07/06/concurrentaffairs/default.aspx
 */

using System;
using System.Threading;
using System.Runtime.InteropServices;
using CyUSB;

namespace CyUSB.Async {

	/// <summary>
	/// Wrapper for the CyUSBDevice that implements .NET-style asynchronous patterns, instead
	/// of the annoying interface that Cypress provides. You pass it a CyUSBEndPoint at construction,
	/// and it will provide nice .NET style asynchronous calls for you to use, which work pretty
	/// nicely. :)
	/// 
	/// It is important to note that this DOES NOT replace the CyUSBDevice class, it is really
	/// just a nice supplement for it. 
	/// </summary>
	public sealed partial class CyUSBAsyncWrapper {

		CyUSBEndPoint endPoint = null;
		IOCompletionCallback transferCallback;


		private CyUSBAsyncWrapper(){}

		public unsafe CyUSBAsyncWrapper(CyUSBEndPoint endPoint) {

			transferCallback = new IOCompletionCallback(TransferCallback);

			this.endPoint = endPoint;

			// don't care about this warning... we have to use an IntPtr, its all we have
#pragma warning disable 0618
			try {
				ThreadPool.BindHandle(endPoint.hDevice);
			} catch (ApplicationException){
				// don't care about this exception
				// it will occur if we try to bind the handle twice, which.. doesn't
				// matter. probably. We'll see how that goes. 
			}
#pragma warning restore 0618

		}
	

		/// <summary>
		/// Begins an asynchronous transfer on the endpoint
		/// </summary>
		/// <param name="buffer">Buffer to read/write</param>
		/// <param name="len">Length to read/write</param>
		public IAsyncResult BeginTransfer(byte[] buffer, int len, AsyncCallback userCallback, object asyncState) {


			// check for stupid mistakes first
			if ( buffer == null )
				throw new ArgumentNullException("buffer");

			if ( buffer.Length < len )
				throw new ArgumentException("Length exceeds the size of the buffer", "len");

			// this doesn't always work, however
			if ( endPoint.hDevice == CyConst.INVALID_HANDLE)
				throw new ArgumentException("The endpoint that this class encompassed is no longer valid");

			// create a new asyncresult
			CyUSBAsyncResult result = new CyUSBAsyncResult();

			// populate its fields
			result.callback = userCallback;
			result.asyncState = asyncState;

			// create the overlapped io struct stuff
			result.overlapped = new Overlapped(0,0,IntPtr.Zero, result);

			// necessary parameters for call
			uint deviceControlCode = 0x22004b;

			bool ret;
			int lpBytesReturned = 0;

			byte [] inputBuffer = new byte[CyConst.SINGLE_XFER_LEN];

			unsafe{

				// pack the buffer and result into the overlapped thing
				NativeOverlapped * ov = result.overlapped.Pack(transferCallback, buffer);

				// pin this down
				fixed ( byte * singleRef = inputBuffer ) {

					// create this thing
					SINGLE_TRANSFER* single_transferPtr = (SINGLE_TRANSFER*)singleRef;
					single_transferPtr->WaitForever = 0;
					single_transferPtr->ucEndpointAddress = endPoint.Address;
					single_transferPtr->IsoPacketLength = 0;
					single_transferPtr->BufferOffset = 0;
					single_transferPtr->BufferLength = 0;

					// finally execute the ioctl
					ret = NativeMethods.DeviceIoControl(endPoint.hDevice, deviceControlCode, single_transferPtr, CyConst.SINGLE_XFER_LEN, buffer, len, out lpBytesReturned, ov);
        
				}
			}
		
			// success
			if (ret) 
				return result;

			Int32 error = Marshal.GetLastWin32Error();

			// io pending.. we expect this to happen
			if (error == 997) 
				return result;

			// other error.. I guess just throw an exception, see what happens
			throw new InvalidOperationException(String.Format("Control failed (code={0})", error));
		}


		// io completion port callback thing
		private unsafe void TransferCallback(uint errorCode, uint numBytes, NativeOverlapped* pOVERLAP) {

			// get the data we need
			Overlapped overlapped = Overlapped.Unpack(pOVERLAP);
			CyUSBAsyncResult result = overlapped.AsyncResult as CyUSBAsyncResult;

			// free this, don't need it anymore
			Overlapped.Free(pOVERLAP);	

			if ( errorCode != 0 ) {

				// some error occurred, tell the user.. not sure what else to do about this
				result.errorCode = errorCode;
			}

			// keep this information around, we need to return it later
			result.bytesTransferred = numBytes;

			// release anyone waiting on this
			result.asyncWaitHandle.Set();

			// if there is a callback, then call it and we're done
			if ( result.callback != null )
				result.callback(result);
		}

		/// <summary>
		/// Ends an asynchronous transfer on the endpoint. Blocks until it finishes. 
		/// </summary>
		public int EndTransfer(IAsyncResult asyncResult) {

			// simple error checking, don't do anything crazy
			if ( asyncResult == null )
				throw new ArgumentNullException("asyncResult");

			// cast the result so we can get its internals
			CyUSBAsyncResult result = asyncResult as CyUSBAsyncResult;

			// wait for the operation to complete -- generally, this shouldn't be necessary
			try {
				result.AsyncWaitHandle.WaitOne();
			} finally {
				// close the handle, we don't need it anymore
				result.AsyncWaitHandle.Close();
			}

			// and... we're done
			result.isCompleted = true;

			// is there an error code? 
			if ( result.errorCode != 0 )
				throw new ApplicationException("USB Transfer failed: error " + result.errorCode.ToString());

			// return the result
			return (int)result.bytesTransferred;

		}

	}

	public sealed class CyUSBAsyncResult : IAsyncResult  {

		internal bool isCompleted = false;

		internal Overlapped overlapped;
		internal object asyncState;
		internal ManualResetEvent asyncWaitHandle = new ManualResetEvent(false);
		internal AsyncCallback callback;

		internal uint bytesTransferred = 0;

		internal uint errorCode = 0;

		#region IAsyncResult Members

		/// <summary>
		/// User-specified state object
		/// </summary>
		public object AsyncState {
			get { return asyncState; }
		}

		public WaitHandle AsyncWaitHandle {
			get { return asyncWaitHandle; }
		}

		/// <summary>
		/// This always returns false
		/// </summary>
		public bool CompletedSynchronously {
			get { return false; }
		}

		public bool IsCompleted {
			get { return isCompleted; }
		}

		#endregion

	}
}
